# -*- coding: utf-8 -*-

"""This module contains classes for IceCube specific detector signal yields,
for a variation of source model and flux model combinations.
"""

import abc
from astropy import units
import numpy as np

import scipy.interpolate

from skyllh.core import (
    multiproc,
)
from skyllh.core.py import (
    classname,
    issequenceof,
)
from skyllh.core.binning import (
    BinningDefinition,
)
from skyllh.core.parameters import (
    ParameterGrid,
)
from skyllh.core.detsigyield import (
    DetSigYield,
    DetSigYieldBuilder,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.source_model import (
    PointLikeSource,
)


class I3DetSigYield(
        DetSigYield,
        metaclass=abc.ABCMeta):
    """Abstract base class for all IceCube specific detector signal yield
    classes. It assumes that sin(dec) binning is required for calculating the
    detector effective area and hence the detector signal yield.
    """

    def __init__(
            self,
            param_names,
            dataset,
            fluxmodel,
            livetime,
            sin_dec_binning,
            **kwargs):
        """Constructor of the IceCube specific detector signal yield base
        class.

        Parameters
        ----------
        param_names : sequence of str
            The sequence of parameter names this detector signal yield depends
            on. These are either fixed or floating parameters.
        dataset : Dataset instance
            The Dataset instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime instance
            The live-time.
        sin_dec_binning : BinningDefinition instance
            The BinningDefinition instance defining the sin(dec) binning.
        """
        super().__init__(
            param_names=param_names,
            dataset=dataset,
            fluxmodel=fluxmodel,
            livetime=livetime,
            **kwargs)

        self.sin_dec_binning = sin_dec_binning

    @property
    def sin_dec_binning(self):
        """The BinningDefinition instance defining the sin(dec) binning.
        """
        return self._sin_dec_binning

    @sin_dec_binning.setter
    def sin_dec_binning(self, bd):
        if not isinstance(bd, BinningDefinition):
            raise TypeError(
                'The sin_dec_binning property must be an instance of '
                f'BinningDefinition! Its current type is {classname(bd)}.')
        self._sin_dec_binning = bd


class I3DetSigYieldBuilder(
        DetSigYieldBuilder,
        metaclass=abc.ABCMeta):
    """Abstract base class for an IceCube specific detector signal yield
    builder class.
    """

    def __init__(
            self,
            sin_dec_binning=None,
            **kwargs,
    ):
        """Constructor of the IceCube specific detector signal yield
        builder class.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition instance
            The instance of BinningDefinition defining the sin(dec) binning.
        """
        super().__init__(**kwargs)

        self.sin_dec_binning = sin_dec_binning

    @property
    def sin_dec_binning(self):
        """The BinningDefinition instance for the sin(dec) binning that should
        be used for computing the sin(dec) dependency of the detector signal
        yield. If None, the binning is supposed to be taken from the Dataset's
        binning definitions.
        """
        return self._sin_dec_binning

    @sin_dec_binning.setter
    def sin_dec_binning(self, binning):
        if (binning is not None) and\
           (not isinstance(binning, BinningDefinition)):
            raise TypeError(
                'The sin_dec_binning property must be None, or an instance of '
                'BinningDefinition! '
                f'Its current type is "{classname(binning)}".')
        self._sin_dec_binning = binning

    def get_sin_dec_binning(self, dataset):
        """Gets the sin(dec) binning definition either as setting from this
        detector signal yield implementation method itself, or from the
        given dataset.
        """
        sin_dec_binning = self.sin_dec_binning
        if sin_dec_binning is None:
            if not dataset.has_binning_definition('sin_dec'):
                raise KeyError(
                    'No binning definition named "sin_dec" is defined in the '
                    f'dataset "{dataset.name}" and no user defined binning '
                    'definition was provided to this detector signal yield '
                    f'builder "{classname(self)}"!')
            sin_dec_binning = dataset.get_binning_definition('sin_dec')
        return sin_dec_binning


class PointLikeSourceI3DetSigYield(
        I3DetSigYield):
    """Abstract base class for all IceCube specific detector signal yield
    classes for point-like sources.
    """

    def __init__(
            self,
            param_names,
            dataset,
            fluxmodel,
            livetime,
            sin_dec_binning,
            **kwargs):
        """Constructor of the IceCube specific detector signal yield base
        class for point-like sources.

        Parameters
        ----------
        param_names : sequence of str
            The sequence of parameter names this detector signal yield depends
            on. These are either fixed or floating parameters.
        dataset : Dataset instance
            The Dataset instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime instance
            The livetime in days or an instance of Livetime.
        sin_dec_binning : BinningDefinition instance
            The BinningDefinition instance defining the sin(dec) binning.
        """
        super().__init__(
            param_names=param_names,
            dataset=dataset,
            fluxmodel=fluxmodel,
            livetime=livetime,
            sin_dec_binning=sin_dec_binning,
            **kwargs)

    def sources_to_recarray(self, sources):
        """Converts the sequence of PointLikeSource sources into a numpy record
        array holding the information of the sources needed for the
        detector signal yield calculation.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source model(s) containing the information of the source(s).

        Returns
        -------
        recarr : numpy record ndarray
            The generated (N_sources,)-shaped 1D numpy record ndarray holding
            the information for each source.
        """
        if isinstance(sources, PointLikeSource):
            sources = [sources]
        if not issequenceof(sources, PointLikeSource):
            raise TypeError(
                'The sources argument must be an instance or a sequence of '
                'instances of PointLikeSource!')

        recarr = np.empty((len(sources),), dtype=[('dec', np.float64)])
        for (i, src) in enumerate(sources):
            recarr['dec'][i] = src.dec

        return recarr


class PointLikeSourceI3DetSigYieldBuilder(
        I3DetSigYieldBuilder,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for all IceCube specific detector signal yield
    builders for point-like sources. All IceCube detector signal
    yield builders require a sin(dec) binning definition for
    the effective area. By default it is taken from the binning definitions
    stored in the dataset, but a user-defined sin(dec) binning can be specified
    if needed.
    """

    def __init__(
            self,
            sin_dec_binning=None,
            **kwargs,
    ):
        """Initializes a new detector signal yield builder object.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition | None
            The BinningDefinition instance defining the sin(dec) binning that
            should be used to compute the sin(dec) dependency of the detector
            effective area. If set to None, the binning will be taken from the
            Dataset binning definitions.
        """
        super().__init__(
            sin_dec_binning=sin_dec_binning,
            **kwargs)


class FixedFluxPointLikeSourceI3DetSigYield(
        PointLikeSourceI3DetSigYield):
    """The detector signal yield class for a point-source with a fixed flux.
    """
    def __init__(
            self,
            param_names,
            dataset,
            fluxmodel,
            livetime,
            sin_dec_binning,
            log_spl_sinDec,
            **kwargs):
        """Constructs an IceCube detector signal yield instance for a
        point-like source with a fixed flux.

        Parameters
        ----------
        param_names : sequence of str
            The sequence of parameter names this detector signal yield depends
            on. These are either fixed or floating parameters.
        dataset : Dataset instance
            The instance of Dataset holding the monte-carlo data this detector
            signal yield is made for.
        fluxmodel : FluxModel instance
            The instance of FluxModel with fixed parameters this detector signal
            yield is made for.
        livetime : float | Livetime instance
            The livetime in days or an instance of Livetime.
        sin_dec_binning : BinningDefinition instance
            The binning definition for sin(dec).
        log_spl_sinDec : scipy.interpolate.InterpolatedUnivariateSpline
            The spline instance representing the log value of the detector
            signal yield as a function of sin(dec).
        """
        super().__init__(
            param_names=param_names,
            dataset=dataset,
            fluxmodel=fluxmodel,
            livetime=livetime,
            sin_dec_binning=sin_dec_binning,
            **kwargs)

        self.log_spl_sinDec = log_spl_sinDec

    @property
    def log_spl_sinDec(self):
        """The :class:`scipy.interpolate.InterpolatedUnivariateSpline` instance
        representing the spline for the log value of the detector signal
        yield as a function of sin(dec).
        """
        return self._log_spl_sinDec

    @log_spl_sinDec.setter
    def log_spl_sinDec(self, spl):
        if not isinstance(spl, scipy.interpolate.InterpolatedUnivariateSpline):
            raise TypeError(
                'The log_spl_sinDec property must be an instance '
                'of scipy.interpolate.InterpolatedUnivariateSpline!')
        self._log_spl_sinDec = spl

    def __call__(self, src_recarray, src_params_recarray=None):
        """Retrieves the detector signal yield for the list of given sources.

        Parameters
        ----------
        src_recarray : numpy record ndarray
            The numpy record ndarray with the field ``dec`` holding the
            declination of the source.
        src_params_recarray : None
            Unused interface argument, because this detector signal yield does
            not depend on any source parameters.

        Returns
        -------
        values : numpy 1d ndarray
            The array with the detector signal yield for each source.
        grads : dict
            This detector signal yield does not depend on any parameters.
            So there are no gradients and the dictionary is empty.
        """
        src_dec = np.atleast_1d(src_recarray['dec'])

        # Create results array.
        values = np.zeros_like(src_dec, dtype=np.float64)

        # Create mask for all source declinations which are inside the
        # declination range.
        mask = (
            (np.sin(src_dec) >= self._sin_dec_binning.lower_edge) &
            (np.sin(src_dec) <= self._sin_dec_binning.upper_edge)
        )

        values[mask] = np.exp(self._log_spl_sinDec(np.sin(src_dec[mask])))

        return (values, {})


class FixedFluxPointLikeSourceI3DetSigYieldBuilder(
        PointLikeSourceI3DetSigYieldBuilder,
        multiproc.IsParallelizable,
):
    """This detector signal yield builder constructs a
    detector signal yield for a fixed flux model, assuming a point-like
    source. This means that the detector signal yield does not depend on
    any source parameters, hence it is only dependent on the detector
    effective area.
    It constructs a one-dimensional spline function in sin(dec), using a
    :class:`scipy.interpolate.InterpolatedUnivariateSpline`.

    This detector signal yield builder works with all flux models.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends soley on the zenith angle, and hence on the
    declination, of the source.
    """

    def __init__(
            self,
            sin_dec_binning=None,
            spline_order_sinDec=2,
            **kwargs,
    ):
        """Creates a new IceCube detector signal yield builder object for a
        fixed flux model. It requires a sinDec binning definition to compute
        the sin(dec) dependency of the detector effective area.
        The construct_detsigyield class method of this builder will create a
        spline function of a given order in logarithmic space of the
        effective area.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition | None
            The BinningDefinition instance which defines the sin(dec) binning.
            If set to None, the binning will be taken from the Dataset binning
            definitions.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        """
        super().__init__(
            sin_dec_binning=sin_dec_binning,
            **kwargs)

        self.spline_order_sinDec = spline_order_sinDec

    @property
    def spline_order_sinDec(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the sin(dec) axis.
        """
        return self._spline_order_sinDec

    @spline_order_sinDec.setter
    def spline_order_sinDec(self, order):
        if not isinstance(order, int):
            raise TypeError(
                'The spline_order_sinDec property must be of type int! '
                f'Its current type is {classname(order)}.')
        self._spline_order_sinDec = order

    def _create_hist(
            self,
            data_sin_true_dec,
            data_true_energy,
            sin_dec_binning,
            weights,
            fluxmodel,
            to_internal_flux_unit_factor,
    ):
        """Creates a histogram of the detector signal yield with the
        given sin(dec) binning for the given flux model.

        Parameters
        ----------
        data_sin_true_dec : instance of numpy.ndarray
            The (N_data,)-shaped numpy.ndarray holding the sin(true_dec) values
            of the monte-carlo events.
        data_true_energy : instance of numpy.ndarray
            The (N_data,)-shaped numpy.ndarray holding the true energy of the
            monte-carlo events.
        sin_dec_binning : instance of BinningDefinition
            The sin(dec) binning definition to use for the histogram.
        weights : 1d ndarray
            The (N_data,)-shaped numpy.ndarray holding the weight factor of
            each monte-carlo event where only the flux value needs to be
            multiplied with in order to get the detector signal yield.
        fluxmodel : instance of FluxModel
            The flux model to get the flux values from.
        to_internal_flux_unit_factor : float
            The conversion factor to convert the flux unit into the internal
            flux unit.

        Returns
        -------
        hist : instance of numpy.ndarray
            The (N_sin_dec_bins,)-shaped numpy.ndarray containing the histogram
            values.
        """
        weights = (
            weights *
            fluxmodel(E=data_true_energy).squeeze() *
            to_internal_flux_unit_factor
        )

        (hist, _) = np.histogram(
            data_sin_true_dec,
            bins=sin_dec_binning.binedges,
            weights=weights,
            density=False)

        # Normalize by solid angle of each bin which is
        # 2*\pi*(\Delta sin(\delta)).
        hist /= (2.*np.pi * np.diff(sin_dec_binning.binedges))

        return hist

    def _create_detsigyield_from_hist(
            self,
            hist,
            sin_dec_binning,
            **kwargs,
    ):
        """Create a single instance of FixedFluxPointLikeSourceI3DetSigYield
        from the given histogram.

        Parameters
        ----------
        hist : instance of numpy.ndarray
            The (N_sin_dec_bins,)-shaped numpy.ndarray holding the normalized
            histogram of the detector signal yield.
        sin_dec_binning : instance of BinningDefinition
            The sin(dec) binning definition to use for the histogram.
        **kwargs
            Additional keyword arguments are passed to the constructor of the
            FixedFluxPointLikeSourceI3DetSigYield class.

        Returns
        -------
        detsigyield : instance of FixedFluxPointLikeSourceI3DetSigYield
            The instance of FixedFluxPointLikeSourceI3DetSigYield for the given
            flux model.
        """
        # Create spline in ln(hist) at the histogram's bin centers.
        log_spl_sinDec = scipy.interpolate.InterpolatedUnivariateSpline(
            sin_dec_binning.bincenters,
            np.log(hist),
            k=self.spline_order_sinDec)

        detsigyield = FixedFluxPointLikeSourceI3DetSigYield(
            param_names=[],
            sin_dec_binning=sin_dec_binning,
            log_spl_sinDec=log_spl_sinDec,
            **kwargs)

        return detsigyield

    def construct_detsigyields(
            self,
            dataset,
            data,
            shgs,
            ppbar=None,
    ):
        """Constructs a set of FixedFluxPointLikeSourceI3DetSigYield instances,
        one for each provided fluxmodel.

        Parameters
        ----------
        dataset : instance of Dataset
            The instance of Dataset holding meta information about the data.
        data : instance of DatasetData
            The instance of DatasetData holding the monte-carlo event data.
            The numpy record ndarray holding the monte-carlo event data must
            contain the following data fields:

            - 'true_dec' : float
                The true declination of the data event.
            - 'true_energy' : float
                The true energy value of the data event.
            - 'mcweight' : float
                The monte-carlo weight of the data event in the unit
                GeV cm^2 sr.

        shgs : sequence of instance of SourceHypoGroup
            The sequence of instance of SourceHypoGroup specifying the
            source hypothesis groups (i.e. flux model) for which the detector
            signal yields should get constructed.
        ppbar : instance of ProgressBar | None
            The optional instance of ProgressBar of the parent progress bar.

        Returns
        -------
        detsigyields : list of instance of FixedFluxPointLikeSourceI3DetSigYield
            The list of instance of FixedFluxPointLikeSourceI3DetSigYield
            providing the detector signal yield function for a point-like source
            with each of the given fixed flux models.
        """
        self.assert_types_of_construct_detsigyield_arguments(
            dataset=dataset,
            data=data,
            shgs=shgs,
            ppbar=ppbar)

        # Calculate conversion factor from the flux model unit into the
        # internal flux unit (usually GeV^-1 cm^-2 s^-1).
        to_internal_flux_unit_factors = [
            shg.fluxmodel.to_internal_flux_unit()
            for shg in shgs
        ]

        to_internal_time_unit_factor = self._cfg.to_internal_time_unit(
            time_unit=units.day
        )

        # Get integrated live-time in days.
        livetime_days = Livetime.get_integrated_livetime(data.livetime)

        # Get the sin(dec) binning definition either as setting from this
        # implementation method, or from the dataset.
        sin_dec_binning = self.get_sin_dec_binning(dataset)

        data_sin_true_dec = np.sin(data.mc['true_dec'])

        # Generate a list of indices that would sort the data according to the
        # sin(true_dec) values. We will sort the MC data according to it,
        # because the histogram creation is much faster (2x) when the
        # to-be-histogrammed values are already sorted.
        sorted_idxs = np.argsort(data_sin_true_dec)

        data_sin_true_dec = np.take(data_sin_true_dec, sorted_idxs)
        data_true_energy = np.take(data.mc['true_energy'], sorted_idxs)
        mc_weight = np.take(data.mc['mcweight'], sorted_idxs)

        weights = (
            mc_weight *
            livetime_days*to_internal_time_unit_factor
        )

        args_list = [
            ((), dict(
                data_sin_true_dec=data_sin_true_dec,
                data_true_energy=data_true_energy,
                sin_dec_binning=sin_dec_binning,
                weights=weights,
                fluxmodel=shg.fluxmodel,
                to_internal_flux_unit_factor=to_internal_flux_unit_factor,
            ))
            for (shg, to_internal_flux_unit_factor) in zip(
                shgs, to_internal_flux_unit_factors)
        ]

        hists = multiproc.parallelize(
            func=self._create_hist,
            args_list=args_list,
            ncpu=multiproc.get_ncpu(
                cfg=self._cfg,
                local_ncpu=self.ncpu),
            ppbar=ppbar,
        )

        detsigyields = [
            self._create_detsigyield_from_hist(
                hist=hist,
                sin_dec_binning=sin_dec_binning,
                dataset=dataset,
                livetime=data.livetime,
                fluxmodel=shg.fluxmodel,
            )
            for (hist, shg) in zip(hists, shgs)
        ]

        return detsigyields

    def construct_detsigyield(
            self,
            dataset,
            data,
            shg,
            ppbar=None,
    ):
        """Constructs a detector signal yield log spline function for the
        given fixed flux model.

        This method calls the :meth:`construct_detsigyiels` method of this
        class.

        Parameters
        ----------
        dataset : instance of Dataset
            The instance of Dataset holding meta information about the data.
        data : instance of DatasetData
            The instance of DatasetData holding the monte-carlo event data.
            The numpy record ndarray holding the monte-carlo event data must
            contain the following data fields:

            - 'true_dec' : float
                The true declination of the data event.
            - 'true_energy' : float
                The true energy value of the data event.
            - 'mcweight' : float
                The monte-carlo weight of the data event in the unit
                GeV cm^2 sr.

        shg : instance of SourceHypoGroup
            The instance of SourceHypoGroup (i.e. sources and flux model) for
            which the detector signal yield should get constructed.
        ppbar : instance of ProgressBar | None
            The optional instance of ProgressBar of the parent progress bar.

        Returns
        -------
        detsigyield : instance of FixedFluxPointLikeSourceI3DetSigYield
            The instance of FixedFluxPointLikeSourceI3DetSigYield providing the
            detector signal yield function for a point-like source with a
            fixed flux.
        """
        detsigyield = self.construct_detsigyields(
            dataset=dataset,
            data=data,
            shgs=[shg],
            ppbar=ppbar,
        )[0]

        return detsigyield

    def get_detsigyield_construction_factory(self):
        """Returns the factory callable for constructing a set of instance of
        FixedFluxPointLikeSourceI3DetSigYield.

        Returns
        -------
        factory : callable
            The factory callable for constructing a set of instance of
            FixedFluxPointLikeSourceI3DetSigYield.
        """
        factory = self.construct_detsigyields
        return factory


class SingleParamFluxPointLikeSourceI3DetSigYield(
        PointLikeSourceI3DetSigYield):
    """The detector signal yield class for a flux that depends on a single
    source parameter.
    """
    def __init__(
            self,
            param_name,
            dataset,
            fluxmodel,
            livetime,
            sin_dec_binning,
            log_spl_sinDec_param,
            **kwargs):
        """Constructs the detector signal yield instance.

        Parameters
        ----------
        param_name : str
            The parameter name this detector signal yield depends
            on. These are either fixed or floating parameter.
        dataset : Dataset instance
            The Dataset instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime instance
            The live-time.
        sin_dec_binning : BinningDefinition instance
            The BinningDefinition instance defining the sin(dec) binning.
        log_spl_sinDec_param : scipy.interpolate.RectBivariateSpline instance
            The 2D spline in sin(dec) and the parameter this detector signal
            yield depends on.
        """
        super().__init__(
            param_names=[param_name],
            dataset=dataset,
            fluxmodel=fluxmodel,
            livetime=livetime,
            sin_dec_binning=sin_dec_binning,
            **kwargs)

        self.log_spl_sinDec_param = log_spl_sinDec_param

    @property
    def log_spl_sinDec_param(self):
        """The :class:`scipy.interpolate.RectBivariateSpline` instance
        representing the spline for the log value of the detector signal
        yield as a function of sin(dec) and the floating parameter.
        """
        return self._log_spl_sinDec_param

    @log_spl_sinDec_param.setter
    def log_spl_sinDec_param(self, spl):
        if not isinstance(spl, scipy.interpolate.RectBivariateSpline):
            raise TypeError(
                'The log_spl_sinDec_param property must be an instance of '
                'scipy.interpolate.RectBivariateSpline! '
                f'Its current type is {classname(spl)}.')
        self._log_spl_sinDec_param = spl

    def __call__(self, src_recarray, src_params_recarray):
        """Retrieves the detector signal yield for the given list of
        sources and their flux parameters.

        Parameters
        ----------
        src_recarray : numpy record ndarray
            The numpy record ndarray with the field ``dec`` holding the
            declination of the source.
        src_params_recarray : (N_sources,)-shaped numpy record ndarray
            The numpy record ndarray containing the parameter values of the
            sources. The parameter values can be different for the different
            sources.
            The record array needs to contain two fields for each source
            parameter, one named <name> with the source's local parameter name
            holding the source's local parameter value, and one named
            <name:gpidx> holding the global parameter index plus one for each
            source value. For values mapping to non-fit parameters, the index
            should be negative.

        Returns
        -------
        values : numpy (N_sources,)-shaped 1D ndarray
            The array with the detector signal yield for each source.
        grads : dict
            The dictionary holding the gradient values for each global floating
            parameter. The key is the global floating parameter index and the
            value is the (N_sources,)-shaped numpy ndarray holding the gradient
            value dY_k/dp_s.
        """
        local_param_name = self.param_names[0]

        src_dec = np.atleast_1d(src_recarray['dec'])
        src_param = src_params_recarray[local_param_name]
        src_param_gp_idxs = src_params_recarray[f'{local_param_name}:gpidx']

        n_sources = len(src_dec)

        # Check for correct input format.
        if not (len(src_param) == n_sources and
                len(src_param_gp_idxs) == n_sources):
            raise RuntimeError(
                f'The length ({len(src_param)}) of the array for the '
                f'source parameter "{local_param_name}" does not match the '
                f'number of sources ({n_sources})!')

        # Calculate the detector signal yield only for the sources for
        # which we actually have detector acceptance. For the other sources,
        # the detector signal yield is zero.
        src_mask = (np.sin(src_dec) >= self._sin_dec_binning.lower_edge) &\
                   (np.sin(src_dec) <= self._sin_dec_binning.upper_edge)

        values = np.zeros((n_sources,), dtype=np.float64)
        values[src_mask] = np.exp(self._log_spl_sinDec_param(
            np.sin(src_dec[src_mask]), src_param[src_mask], grid=False))

        # Determine the number of global parameters the local parameter is
        # made of.
        gfp_idxs = np.unique(src_param_gp_idxs)
        gfp_idxs = gfp_idxs[gfp_idxs > 0] - 1

        # Calculate the gradients for each global fit parameter.
        grads = dict()
        for gfp_idx in gfp_idxs:
            # Create the gradient array of shape (n_sources,). This could be
            # a masked array to save memory, when there are many sources and
            # global fit parameters.
            grads[gfp_idx] = np.zeros((n_sources,), dtype=np.float64)

            # Create a mask to select the sources that depend on the global
            # fit parameter with index gfp_idx.
            gfp_src_mask = (src_param_gp_idxs == gfp_idx+1)

            # m is a (n_sources,)-shaped ndarray, which selects only sources
            # that have detector exceptance and depend on the global fit
            # parameter gfp_idx.
            m = src_mask & gfp_src_mask

            grads[gfp_idx][m] = values[m] * self._log_spl_sinDec_param(
                np.sin(src_dec[m]), src_param[m], grid=False, dy=1)

        return (values, grads)


class SingleParamFluxPointLikeSourceI3DetSigYieldBuilder(
        PointLikeSourceI3DetSigYieldBuilder,
        multiproc.IsParallelizable,
):
    """This detector signal yield builder constructs a
    detector signal yield for a variable flux model with a single parameter,
    assuming a point-like source.
    It constructs a two-dimensional spline function in sin(dec) and the
    parameter, using a :class:`scipy.interpolate.RectBivariateSpline`.
    Hence, the detector signal yield can vary with the declination and the
    parameter of the flux model.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends soley on the zenith angle, and hence on the
    declination, of the source.
    """
    def __init__(
            self,
            param_grid,
            sin_dec_binning=None,
            spline_order_sinDec=2,
            spline_order_param=2,
            ncpu=None,
            **kwargs,
    ):
        """Creates a new IceCube detector signal yield builder instance for a
        flux model with a single parameter.
        It requires a sinDec binning definition to compute the sin(dec)
        dependency of the detector effective area, and a parameter grid to
        compute the parameter dependency of the detector signal yield.

        Parameters
        ----------
        param_grid : instance of ParameterGrid
            The instance of ParameterGrid which defines the grid of the
            parameter values. The name of the parameter is defined via the name
            property of the ParameterGrid instance.
        sin_dec_binning : instance of BinningDefinition | None
            The instance of BinningDefinition which defines the sin(dec)
            binning. If set to None, the sin(dec) binning will be taken from the
            dataset's binning definitions.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        spline_order_gamma : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the gamma axis.
            The default is 2.
        ncpu : int | None
            The number of CPUs to utilize. If set to ``None``, global setting
            will take place.
        """
        super().__init__(
            sin_dec_binning,
            ncpu=ncpu,
            **kwargs)

        self.param_grid = param_grid
        self.spline_order_sinDec = spline_order_sinDec
        self.spline_order_param = spline_order_param

    @property
    def param_grid(self):
        """The ParameterGrid instance for the parameter grid that should be used
        for computing the parameter dependency of the detector signal yield.
        """
        return self._param_grid

    @param_grid.setter
    def param_grid(self, grid):
        if not isinstance(grid, ParameterGrid):
            raise TypeError(
                'The param_grid property must be an instance of '
                'ParameterGrid! '
                f'Its current type is {classname(grid)}.')
        self._param_grid = grid

    @property
    def spline_order_sinDec(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the sin(dec) axis.
        """
        return self._spline_order_sinDec

    @spline_order_sinDec.setter
    def spline_order_sinDec(self, order):
        if not isinstance(order, int):
            raise TypeError(
                'The spline_order_sinDec property must be of type int! '
                f'Its current type is {classname(order)}.')
        self._spline_order_sinDec = order

    @property
    def spline_order_param(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the parameter axis.
        """
        return self._spline_order_param

    @spline_order_param.setter
    def spline_order_param(self, order):
        if not isinstance(order, int):
            raise TypeError(
                'The spline_order_param property must be of type int! '
                f'Its current type is {classname(order)}.')
        self._spline_order_param = order

    def construct_detsigyield(
            self,
            dataset,
            data,
            shg,
            ppbar=None,
    ):
        """Constructs a detector signal yield 2-dimensional log spline
        function for the given flux model with varying parameter values.

        Parameters
        ----------
        dataset : instance of Dataset
            The instance of Dataset holding the sin(dec) binning definition.
        data : instance of DatasetData
            The instance of DatasetData holding the monte-carlo event data.
            The numpy record array for the monte-carlo data of the dataset must
            contain the following data fields:

            ``'true_dec'`` : float
                The true declination of the data event.
            ``'mcweight'`` : float
                The monte-carlo weight of the data event in the unit
                GeV cm^2 sr.
            ``'true_energy'`` : float
                The true energy value of the data event.

        shg : instance of SourceHypoGroup
            The instance of SourceHypoGroup for which the detector signal yield
            should get constructed.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : instance of SingleParamFluxPointLikeSourceI3DetSigYield
            The I3DetSigYield instance for a point-like source with a flux model
            of a single parameter.
        """
        self.assert_types_of_construct_detsigyield_arguments(
            dataset=dataset,
            data=data,
            shgs=shg,
            ppbar=ppbar)

        # Get integrated live-time in days.
        livetime_days = Livetime.get_integrated_livetime(data.livetime)

        # Get the sin(dec) binning definition either as setting from this
        # implementation method, or from the dataset.
        sin_dec_binning = self.get_sin_dec_binning(dataset)

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit GeV^-1 cm^-2 s^-1.
        to_internal_flux_unit_factor = shg.fluxmodel.to_internal_flux_unit()

        to_internal_time_unit_factor = self._cfg.to_internal_time_unit(
            time_unit=units.day
        )

        # Define a function that creates a detector signal yield histogram
        # along sin(dec) for a given flux model, i.e. for given spectral index,
        # gamma.
        def _create_hist(
                data_sin_true_dec,
                data_true_energy,
                sin_dec_binning,
                weights,
                fluxmodel,
                to_internal_flux_unit_factor,
        ):
            """Creates a histogram of the detector signal yield with the
            given sin(dec) binning.

            Parameters
            ----------
            data_sin_true_dec : 1d ndarray
                The sin(true_dec) values of the monte-carlo events.
            data_true_energy : 1d ndarray
                The true energy of the monte-carlo events.
            sin_dec_binning : BinningDefinition
                The sin(dec) binning definition to use for the histogram.
            weights : 1d ndarray
                The weight factors of each monte-carlo event where only the
                flux value needs to be multiplied with in order to get the
                detector signal yield.
            fluxmodel : FluxModel
                The flux model to get the flux values from.
            to_internal_flux_unit_factor : float
                The conversion factor to convert the flux unit into the internal
                flux unit.

            Returns
            -------
            h : 1d ndarray
                The numpy array containing the histogram values.
            """
            weights = (
                weights *
                fluxmodel(E=data_true_energy).squeeze() *
                to_internal_flux_unit_factor
            )

            (h, edges) = np.histogram(
                data_sin_true_dec,
                bins=sin_dec_binning.binedges,
                weights=weights,
                density=False)

            return h

        data_sin_true_dec = np.sin(data.mc['true_dec'])

        # Generate a list of indices that would sort the data according to the
        # sin(true_dec) values. We will sort the MC data according to it,
        # because the histogram creation is much faster (2x) when the
        # to-be-histogrammed values are already sorted.
        sorted_idxs = np.argsort(data_sin_true_dec)

        data_sin_true_dec = np.take(data_sin_true_dec, sorted_idxs)
        data_true_energy = np.take(data.mc['true_energy'], sorted_idxs)

        weights = (
            np.take(data.mc['mcweight'], sorted_idxs) *
            livetime_days*to_internal_time_unit_factor
        )

        # Make a copy of the parameter grid and extend the grid by one bin on
        # each side.
        param_grid = self._param_grid.copy()
        param_grid.add_extra_lower_and_upper_bin()

        # Construct the arguments for the hist function to be used in the
        # multiproc.parallelize function.
        args_list = [
            (
                (
                    data_sin_true_dec,
                    data_true_energy,
                    sin_dec_binning,
                    weights,
                    shg.fluxmodel.copy({param_grid.name: param_val}),
                    to_internal_flux_unit_factor,
                ),
                {}
            )
            for param_val in param_grid.grid
        ]
        h = np.vstack(
            multiproc.parallelize(
                _create_hist, args_list, self.ncpu, ppbar=ppbar)).T

        # Normalize by solid angle of each bin along the sin(dec) axis.
        # The solid angle is given by 2*\pi*(\Delta sin(\delta)).
        h /= (2.*np.pi * np.diff(sin_dec_binning.binedges)).reshape(
            (sin_dec_binning.nbins, 1))

        # Create the 2D spline.
        log_spl_sinDec_param = scipy.interpolate.RectBivariateSpline(
            sin_dec_binning.bincenters,
            param_grid.grid,
            np.log(h),
            kx=self.spline_order_sinDec,
            ky=self.spline_order_param,
            s=0)

        detsigyield = SingleParamFluxPointLikeSourceI3DetSigYield(
            param_name=self._param_grid.name,
            dataset=dataset,
            fluxmodel=shg.fluxmodel,
            livetime=data.livetime,
            sin_dec_binning=sin_dec_binning,
            log_spl_sinDec_param=log_spl_sinDec_param)

        return detsigyield



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

import numpy as np

from skyllh.datasets.i3 import (
    PublicData_10y_ps,
)


def create_dataset_collection(
        cfg,
        base_path=None,
        sub_path_fmt=None,
):
    """Defines the dataset collection for IceCube's 10-year
    point-source public data, which is available at
    http://icecube.wisc.edu/data-releases/20210126_PS-IC40-IC86_VII.zip, and
    adds monte-carlo files.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    base_path : str | None
        The base path of the data files. The actual path of a data file is
        assumed to be of the structure <base_path>/<sub_path>/<file_name>.
        If None, use the default path ``cfg['repository']['base_path']``.
    sub_path_fmt : str | None
        The sub path format of the data files of the public data sample.
        If None, use the default sub path format
        'icecube_10year_ps'.

    Returns
    -------
    dsc : DatasetCollection
        The dataset collection containing all the seasons as individual
        I3Dataset objects.
    """
    dsc = PublicData_10y_ps.create_dataset_collection(
        cfg=cfg,
        base_path=base_path,
        sub_path_fmt=sub_path_fmt)

    dsc.description += """
    This dataset collection has monte-carlo (MC) files defined for each dataset.
    These MC files are not part of the original data release and need to be
    generated by the user.
    """

    (
        IC40,
        IC59,
        IC79,
        IC86_I,
        IC86_II,
        IC86_III,
        IC86_IV,
        IC86_V,
        IC86_VI,
        IC86_VII,
        IC86_II_VII,
    ) = dsc[
        'IC40',
        'IC59',
        'IC79',
        'IC86_I',
        'IC86_II',
        'IC86_III',
        'IC86_IV',
        'IC86_V',
        'IC86_VI',
        'IC86_VII',
        'IC86_II-VII',
    ]
    IC40.mc_pathfilename_list = 'sim/IC40_MC.npy'
    IC59.mc_pathfilename_list = 'sim/IC59_MC.npy'
    IC79.mc_pathfilename_list = 'sim/IC79_MC.npy'
    IC86_I.mc_pathfilename_list = 'sim/IC86_I_MC.npy'
    IC86_II.mc_pathfilename_list = 'sim/IC86_II-VII_MC.npy'
    IC86_III.mc_pathfilename_list = IC86_II.mc_pathfilename_list
    IC86_IV.mc_pathfilename_list = IC86_II.mc_pathfilename_list
    IC86_V.mc_pathfilename_list = IC86_II.mc_pathfilename_list
    IC86_VI.mc_pathfilename_list = IC86_II.mc_pathfilename_list
    IC86_VII.mc_pathfilename_list = IC86_II.mc_pathfilename_list
    IC86_II_VII.mc_pathfilename_list = IC86_II.mc_pathfilename_list

    def add_time(data):
        mc = data.mc
        mc.append_field('time', np.repeat(0, len(mc)))

    def add_azimuth_and_zenith(data):
        mc = data.mc
        mc.append_field('azi', np.repeat(0, len(mc)))
        mc.append_field('zen', np.repeat(0, len(mc)))

    dsc.add_data_preparation(add_time)
    dsc.add_data_preparation(add_azimuth_and_zenith)

    return dsc



# -*- coding: utf-8 -*-

import abc
import numpy as np

import scipy.signal
import scipy.stats

from skyllh.core.py import issequenceof

# Define a constant that can be used when specifying a histogram axis as
# unsmooth, i.e. no smoothing should be applied along that axis.
UNSMOOTH_AXIS = np.ones(1)


class HistSmoothingMethod(
        object,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for implementing a histogram smoothing method.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def smooth(self, h):
        """This method is supposed to smooth the given histogram h.

        Parameters
        ----------
        h : N-dimensional ndarray
            The ndarray holding histogram bin values.

        Returns
        -------
        smoothed_h : N-dimensional ndarray
            The array holding the smoothed histogram bin values.
        """
        pass


class NoHistSmoothingMethod(
        HistSmoothingMethod,
):
    """This class implements a no-shoothing histogram method.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def smooth(self, h):
        """Does not perform any smoothing and just returns the input histogram.

        Parameters
        ----------
        h : N-dimensional ndarray
            The ndarray holding histogram bin values.

        Returns
        -------
        h : N-dimensional ndarray
            The input histogram array.
        """
        return h


class NeighboringBinHistSmoothingMethod(
        HistSmoothingMethod,
):
    """This class implements a smoothing algorithm that smoothes a histogram
    based on the neighboring bins.
    """
    def __init__(
            self,
            axis_kernel_arrays,
            **kwargs,
    ):
        """Constructs a new neighboring bin histogram smoothing method.

        Parameters
        ----------
        axis_kernel_arrays: sequence of 1D ndarrays
            The sequence of smoothing kernel arrays, one for each axis. If an
            axis should not get smoothed, the UNSMOOTH_AXIS constant should be
            used for that axis' smoothing kernel array.
        """
        super().__init__(**kwargs)

        if not issequenceof(axis_kernel_arrays, np.ndarray):
            raise TypeError(
                'The axis_kernel_arrays argument must be a sequence of '
                'numpy.ndarray instances!')

        self._ndim = len(axis_kernel_arrays)

        # Construct the smoothing kernel k used by the smooth method.
        # k is a N-dimensional ndarray. It defines which neighboring bin values
        # of the histogram will contribute how much to the central bin value.
        self._k = np.prod(
            np.meshgrid(*axis_kernel_arrays, indexing='ij'), axis=0)

    @property
    def ndim(self):
        """(read-only) The dimensionality of the histogram this smoothing
        instances is made for.
        """
        return self._ndim

    def smooth(self, h):
        """Smoothes the given histogram array h with the internal kernel array
        k. Both arrays must have the same dimensionality. The shape
        values of k must be smaller than or equal to the shape values of h.

        Parameters
        ----------
        h : N-dimensional ndarray
            The ndarray holding histogram bin values.

        Returns
        -------
        smoothed_h : N-dimensional ndarray.
        """
        if h.ndim != self._ndim:
            raise ValueError(
                'The ndarrays of argument h and k must have the same '
                f'dimensionality! Currently they are {h.ndim:d} and '
                f'{self._ndim:d}, respectively.')
        for d in range(h.ndim):
            if self._k.shape[d] > h.shape[d]:
                raise ValueError(
                    f'The shape value ({self._k.shape[d]:d}) of dimension '
                    f'{d:d} of ndarray k must be smaller than or equal to the '
                    f'shape value ({h.shape[d]:d}) of dimension {d:d} of '
                    'ndarray h!')

        norm = scipy.signal.convolve(np.ones_like(h), self._k, mode="same")
        smoothed_h = scipy.signal.convolve(h, self._k, mode="same") / norm

        return smoothed_h


class SmoothingFilter(
        object):
    """This class provides a base class for a histogram smoothing filter. It
    provides an axis kernel array that defines how many neighboring bins of a
    histogram bin should be used to smooth that histogram bin.
    """
    def __init__(
            self,
            axis_kernel_array,
            **kwargs):
        super().__init__(**kwargs)

        self.axis_kernel_array = axis_kernel_array

    @property
    def axis_kernel_array(self):
        """The kernel array for a histogram axis.
        """
        return self._axis_kernel_array

    @axis_kernel_array.setter
    def axis_kernel_array(self, arr):
        if not isinstance(arr, np.ndarray):
            raise TypeError(
                'The axis_kernel_array property must be an instance of '
                'numpy.ndarray!')
        self._axis_kernel_array = arr


class BlockSmoothingFilter(
        SmoothingFilter,
):
    """This class defines the histogram smoothing filter for smoothing a
    histogram via a block kernel function. The half-width of that
    block is specified via the nbins argument.
    """
    def __init__(
            self,
            nbins,
            **kwargs
    ):
        """Creates a new BlockSmoothingFilter instance.

        Parameters
        ----------
        nbins : int
            The number of neighboring bins into one direction of a histogram
            bin, which should be used to smooth that histogram bin.
        """
        if not isinstance(nbins, int):
            raise TypeError(
                'The nbins argument must be of type int!')
        if nbins <= 0:
            raise ValueError(
                'The nbins argument must be greater zero!')

        arr = np.ones(2*nbins + 1, dtype=np.float64)

        super().__init__(arr, **kwargs)


class GaussianSmoothingFilter(
        SmoothingFilter,
):
    """This class defines the histogram smoothing filter for smoothing a
    histogram via a Gaussian kernel function. The width of that Gaussian is
    approximately one standard deviation, spread over nbins on each side of the
    central histogram bin.
    """
    def __init__(
            self,
            nbins,
            **kwargs,
    ):
        """Creates a new GaussianSmoothingFilter instance.

        Parameters
        ----------
        nbins : int
            The number of neighboring bins into one direction of a histogram
            bin, which should be used to smooth that histogram bin.
        """
        if not isinstance(nbins, int):
            raise TypeError(
                'The nbins argument must be of type int!')
        if nbins <= 0:
            raise ValueError(
                'The nbins argument must be greater zero!')

        val = 1.6635
        r = np.linspace(-val, val, 2*nbins + 1)
        arr = scipy.stats.norm.pdf(r)

        super().__init__(arr, **kwargs)



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

"""This module defines the base class for any model class used in SkyLLH.
"""

from skyllh.core.py import (
    NamedObjectCollection,
    issequenceof,
    str_cast,
    typename,
)


class Model(
        object):
    """This class provides a base class for all model classes used in SkyLLH.
    Models could be for instance source models or background models.
    """
    def __init__(
            self,
            name=None,
            **kwargs):
        """Creates a new Model instance.

        Parameters
        ----------
        name : str | None
            The name of the model. If set to `None`, the id of the object is
            taken as name.
        """
        super().__init__(
            **kwargs)

        if name is None:
            name = self.id

        self.name = name

    @property
    def name(self):
        """The name of the model.
        """
        return self._name

    @name.setter
    def name(self, name):
        name = str_cast(
            name,
            'The name property must be castable to type str!')
        self._name = name

    @property
    def id(self):
        """(read-only) The ID of the model. It's an integer generated with
        Python's `id` function. Hence, it's related to the memory address
        of the object.
        """
        return id(self)


class ModelCollection(
        NamedObjectCollection):
    """This class describes a collection of Model instances. It can be
    used to group several models into a single object.
    """
    @staticmethod
    def cast(
            obj,
            errmsg=None,
            **kwargs):
        """Casts the given object to a ModelCollection object.
        If the cast fails, a TypeError with the given error message is raised.

        Parameters
        ----------
        obj : Model instance | sequence of Model instances |
                ModelCollection | None
            The object that should be casted to ModelCollection.
            If set to None, an empty ModelCollection is created.
        errmsg : str | None
            The error message if the cast fails.
            If set to None, a generic error message will be used.

        Additional keyword arguments
        ----------------------------
        Additional keyword arguments are passed to the constructor of the
        ModelCollection class.

        Raises
        ------
        TypeError
            If the cast failed.

        Returns
        -------
        model_collection : instance of ModelCollection
            The created ModelCollection instance. If `obj` is already a
            ModelCollection instance, it will be returned.
        """
        if obj is None:
            return ModelCollection(
                models=None, model_type=Model, **kwargs)

        if isinstance(obj, Model):
            return ModelCollection(
                models=[obj], model_type=Model, **kwargs)

        if isinstance(obj, ModelCollection):
            return obj

        if issequenceof(obj, Model):
            return ModelCollection(
                models=obj, model_type=Model, **kwargs)

        if errmsg is None:
            errmsg = (f'Cast of object "{str(obj)}" of type '
                      f'"{typename(type(obj))}" to ModelCollection failed!')
        raise TypeError(errmsg)

    def __init__(
            self,
            models=None,
            model_type=None,
            **kwargs):
        """Creates a new Model collection. The type of the model instances this
        collection holds can be restricted, by setting the model_type argument.

        Parameters
        ----------
        models : sequence of model_type instances | None
            The sequence of models this collection should be initalized with.
        model_type : type | None
            The type of the model. It must be a subclass of class ``Model``.
            If set to None (default), Model will be used.
        """
        if model_type is None:
            model_type = Model
        if not issubclass(model_type, Model):
            raise TypeError(
                'The model_type argument must be a subclass of Model!')

        super().__init__(
            objs=models,
            obj_type=model_type,
            **kwargs)

    @property
    def model_type(self):
        """(read-only) The type of the model.
        """
        return self.obj_type

    @property
    def models(self):
        """(read-only) The list of models of type `model_type`.
        """
        return self.objects


class DetectorModel(Model):
    """This class provides a base class for a detector model. It can be used
    in combination with the ParameterModelMapper class.
    """
    def __init__(self, name, **kwargs):
        """Creates a new DetectorModel instance.

        Parameters
        ----------
        name : str
            The name of the detector model.
        """
        super().__init__(
            name=name,
            **kwargs)



# -*- coding: utf-8 -*-

import abc

import numpy as np

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.interpolate import (
    GridManifoldInterpolationMethod,
    Parabola1DGridManifoldInterpolationMethod,
)
from skyllh.core.parameters import (
    ParameterModelMapper,
)
from skyllh.core.pdf import (
    PDFSet,
    IsBackgroundPDF,
    IsSignalPDF,
)
from skyllh.core.py import (
    classname,
    float_cast,
    int_cast,
    issequence,
    issequenceof,
)
from skyllh.core.services import (
    SrcDetSigYieldWeightsService,
)
from skyllh.core.timing import (
    TaskTimer,
)


class PDFRatio(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for a signal over background PDF ratio class.
    It defines the interface of a signal over background PDF ratio class.
    """

    def __init__(
            self,
            sig_param_names=None,
            bkg_param_names=None,
            **kwargs,
    ):
        """Creates a new PDFRatio instance.

        Parameters
        ----------
        sig_param_names : sequence of str | str | None
            The sequence of signal parameter names this PDFRatio instance is a
            function of.
        bkg_param_names : sequence of str | str | None
            The sequence of background parameter names this PDFRatio instance
            is a function of.
        """
        super().__init__(**kwargs)

        self.sig_param_names = sig_param_names
        self.bkg_param_names = bkg_param_names

    @property
    def n_params(self):
        """(read-only) The number of parameters the PDF ratio depends on.
        This is the sum of signal and background parameters.
        """
        return self.n_sig_params + self.n_bkg_params

    @property
    def param_names(self):
        """(read-only) The list of parameter names this PDF ratio is a
        function of. This is the superset of signal and background parameter
        names.
        """
        return list(
            set(list(self._sig_param_names) + list(self._bkg_param_names)))

    @property
    def n_sig_params(self):
        """(read-only) The number of signal parameters the PDF ratio depends
        on.
        """
        return len(self._sig_param_names)

    @property
    def n_bkg_params(self):
        """(read-only) The number of background parameters the PDF ratio depends
        on.
        """
        return len(self._bkg_param_names)

    @property
    def sig_param_names(self):
        """The list of signal parameter names this PDF ratio is a function of.
        """
        return self._sig_param_names

    @sig_param_names.setter
    def sig_param_names(self, names):
        if names is None:
            names = []
        if not issequence(names):
            names = [names]
        if not issequenceof(names, str):
            raise TypeError(
                'The sig_param_names property must be a sequence of str '
                'instances!')
        self._sig_param_names = names

    @property
    def bkg_param_names(self):
        """The list of background parameter names this PDF ratio is a function
        of.
        """
        return self._bkg_param_names

    @bkg_param_names.setter
    def bkg_param_names(self, names):
        if names is None:
            names = []
        if not issequence(names):
            names = [names]
        if not issequenceof(names, str):
            raise TypeError(
                'The bkg_param_names property must be a sequence of str '
                'instances!')
        self._bkg_param_names = names

    @abc.abstractmethod
    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs,
    ):
        """Initializes the PDFRatio instance for a new trial. This method can
        be utilized to pre-calculate PDFRatio values that do not depend on any
        fit parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial data.
        tl : instance of TimeLord
            The optional instance of TimeLord to measure timing information.
        """
        pass

    @abc.abstractmethod
    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None,
    ):
        """Retrieves the PDF ratio value for each given trial data events (and
        sources), given the given set of parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy record ndarray | None
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : instance of ndarray
            The (N_values,)-shaped 1d numpy ndarray of float holding the PDF
            ratio value for each trial event and source.
        """
        pass

    @abc.abstractmethod
    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None,
    ):
        """Retrieves the PDF ratio gradient for the global fit parameter
        ``fitparam_id`` for each trial data event and source, given the given
        set of parameters ``src_params_recarray`` for each source.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio gradient values should get calculated.
        src_params_recarray : instance of numpy structured ndarray
            The (N_sources,)-shaped numpy structured ndarray holding the
            parameter names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        fitparam_id : int
            The ID of the global fit parameter for which the gradient should
            get calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        gradient : instance of ndarray | 0
            The (N_values,)-shaped 1d numpy ndarray of float holding the PDF
            ratio gradient value for each source and trial data event.
            If the PDF ratio does not depend on the given global fit parameter,
            0 should be returned.
        """
        pass

    def __mul__(self, other):
        """Implements the mathematical operation ``new = self * other``, where
        ``other`` is an instance of PDFRatio. It creates an instance of
        PDFRatioProduct holding the two PDFRatio instances.
        """
        return PDFRatioProduct(self, other, cfg=self.cfg)


class PDFRatioProduct(
        PDFRatio,
):
    """This is the mathematical product of two PDFRatio instances, which is a
    PDFRatio instance again.
    """
    def __init__(
            self,
            pdfratio1,
            pdfratio2,
            **kwargs,
    ):
        """Creates a new PDFRatioProduct instance representing the product of
        two PDFRatio instances.
        """
        self.pdfratio1 = pdfratio1
        self.pdfratio2 = pdfratio2

        sig_param_names = set(
            list(pdfratio1.sig_param_names) + list(pdfratio2.sig_param_names))
        bkg_param_names = set(
            list(pdfratio1.bkg_param_names) + list(pdfratio2.bkg_param_names))

        super().__init__(
            sig_param_names=sig_param_names,
            bkg_param_names=bkg_param_names,
            **kwargs)

    @property
    def pdfratio1(self):
        """The first PDFRatio instance in the muliplication
        ``pdfratio1 * pdfratio2``.
        """
        return self._pdfratio1

    @pdfratio1.setter
    def pdfratio1(self, pdfratio):
        if not isinstance(pdfratio, PDFRatio):
            raise TypeError(
                'The pdfratio1 property must be an instance of PDFRatio!')
        self._pdfratio1 = pdfratio

    @property
    def pdfratio2(self):
        """The second PDFRatio instance in the muliplication
        ``pdfratio1 * pdfratio2``.
        """
        return self._pdfratio2

    @pdfratio2.setter
    def pdfratio2(self, pdfratio):
        if not isinstance(pdfratio, PDFRatio):
            raise TypeError(
                'The pdfratio2 property must be an instance of PDFRatio!')
        self._pdfratio2 = pdfratio

    def initialize_for_new_trial(
            self,
            **kwargs):
        """Initializes the PDFRatioProduct instance for a new trial.
        It calls the
        :meth:`~skyllh.core.pdfratio.PDFRatio.initialize_for_new_trial` method
        of each of the two :class:`~skyllh.core.pdfratio.PDFRatio` instances.
        """
        self._pdfratio1.initialize_for_new_trial(**kwargs)
        self._pdfratio2.initialize_for_new_trial(**kwargs)

    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None,
    ):
        """Retrieves the PDF ratio product value for each trial data
        event and source, given the given set of parameters for all sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy record ndarray
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : instance of ndarray
            The (N_values,)-shaped 1d numpy ndarray of float holding the product
            of the PDF ratio values for each trial event and source.
            The PDF ratio product value for each trial event.
        """
        r1 = self._pdfratio1.get_ratio(
            tdm=tdm,
            src_params_recarray=src_params_recarray,
            tl=tl)

        r2 = self._pdfratio2.get_ratio(
            tdm=tdm,
            src_params_recarray=src_params_recarray,
            tl=tl)

        return r1 * r2

    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None,
    ):
        """Retrieves the PDF ratio product gradient for the global fit parameter
        with parameter ID ``fitparam_id`` for each trial data event and source,
        given the set of parameters ``src_params_recarray`` for all sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy record ndarray | None
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        fitparam_id : int
            The ID of the global fit parameter for which the gradient should
            get calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        gradient : instance of ndarray | 0
            The (N_values,)-shaped 1d numpy ndarray of float holding the PDF
            ratio gradient value for each trial event and source. If none of the
            two PDFRatio instances depend on the given global fit parameter, the
            scalar value ``0`` is returned.
        """
        r1_depends_on_fitparam =\
            ParameterModelMapper.is_global_fitparam_a_local_param(
                fitparam_id=fitparam_id,
                params_recarray=src_params_recarray,
                local_param_names=self._pdfratio1.param_names)

        r2_depends_on_fitparam =\
            ParameterModelMapper.is_global_fitparam_a_local_param(
                fitparam_id=fitparam_id,
                params_recarray=src_params_recarray,
                local_param_names=self._pdfratio2.param_names)

        if r1_depends_on_fitparam:
            r2 = self._pdfratio2.get_ratio(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                tl=tl)

            r1_grad = self._pdfratio1.get_gradient(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                fitparam_id=fitparam_id,
                tl=tl)

        if r2_depends_on_fitparam:
            r1 = self._pdfratio1.get_ratio(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                tl=tl)

            r2_grad = self._pdfratio2.get_gradient(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                fitparam_id=fitparam_id,
                tl=tl)

        if r1_depends_on_fitparam and r2_depends_on_fitparam:
            gradient = r1 * r2_grad
            gradient += r1_grad * r2
        elif r1_depends_on_fitparam:
            gradient = r1_grad * r2
        elif r2_depends_on_fitparam:
            gradient = r1 * r2_grad
        else:
            gradient = 0

        return gradient


class SourceWeightedPDFRatio(
        PDFRatio):
    r"""This class provides the calculation of a source weighted PDF ratio for
    multiple sources:

    .. math::

        \mathcal{R}_i(\vec{p}_{\mathrm{s}}) = \frac{1}{A(\vec{p}_{\mathrm{s}})}
            \sum_{k=1}^{K} a_k(\vec{p}_{\mathrm{s}_k}) \mathcal{R}_{i,k}
            (\vec{p}_{\mathrm{s}_k})

    """
    def __init__(
            self,
            dataset_idx,
            src_detsigyield_weights_service,
            pdfratio,
            **kwargs):
        """Creates a new SourceWeightedPDFRatio instance.

        Parameters
        ----------
        dataset_idx : int
            The index of the dataset. It is used to access the source detector
            signal yield weight.
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            The instance of SrcDetSigYieldWeightsService providing the source
            detector signal yield weights, i.e. the product of the theoretical
            source weight with the detector signal yield.
        pdfratio : instance of PDFRatio
            The instance of PDFRatio providing the PDF ratio values and
            derivatives.
        """
        if not isinstance(pdfratio, PDFRatio):
            raise TypeError(
                'The pdfratio argument must be an instance of PDFRatio! '
                f'Its current type is {classname(pdfratio)}.')
        self._pdfratio = pdfratio

        super().__init__(
            sig_param_names=self._pdfratio.sig_param_names,
            bkg_param_names=self._pdfratio.bkg_param_names,
            **kwargs)

        self._dataset_idx = int_cast(
            dataset_idx,
            'The dataset_idx argument must be castable to type int!')

        if not isinstance(
                src_detsigyield_weights_service,
                SrcDetSigYieldWeightsService):
            raise TypeError(
                'The src_detsigyield_weights_service argument must be an '
                'instance of type SrcDetSigYieldWeightsService! '
                'Its current type is '
                f'{classname(src_detsigyield_weights_service)}.')
        self._src_detsigyield_weights_service = src_detsigyield_weights_service

        self._cache_R_ik = None
        self._cache_R_i = None

    @property
    def dataset_idx(self):
        """(read-only) The index of the dataset for which this
        SourceWeightedPDFRatio instance is made.
        """
        return self._dataset_idx

    @property
    def src_detsigyield_weights_service(self):
        """(read-only) The instance of SrcDetSigYieldWeightsService providing
        the source detector signal yield weights, i.e. the product of the
        theoretical source weight with the detector signal yield.
        """
        return self._src_detsigyield_weights_service

    @property
    def pdfratio(self):
        """(read-only) The PDFRatio instance that is used to calculate the
        source weighted PDF ratio.
        """
        return self._pdfratio

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Initializes the PDFRatio instance for a new trial. It calls the
        :meth:`~skyllh.core.pdfratio.PDFRatio.initialize_for_new_trial` method
        of the :class:`~skyllh.core.pdfratio.PDFRatio` instance.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial data.
        tl : instance of TimeLord
            The optional instance of TimeLord to measure timing information.
        """
        self._pdfratio.initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)

    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None):
        """Retrieves the PDF ratio value for each given trial data events (and
        sources), given the given set of parameters.

        Note:

            This method uses the source detector signal yield weights service.
            Hence, the
            :meth:`skyllh.core.weights.SrcDetSigYieldWeightsService.calculate`
            method needs to be called prior to calling this method.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy record ndarray
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : instance of ndarray
            The (N_selected_events,)-shaped 1d numpy ndarray of float holding
            the PDF ratio value for each selected trial data event.
        """
        (a_jk, a_jk_grads) = self._src_detsigyield_weights_service.get_weights()
        a_k = a_jk[self._dataset_idx]

        n_sources = len(a_k)
        n_sel_events = tdm.n_selected_events

        A = np.sum(a_k)

        R_ik = self._pdfratio.get_ratio(
            tdm=tdm,
            src_params_recarray=src_params_recarray,
            tl=tl)
        # The R_ik ndarray is (N_values,)-shaped.

        R_i = np.zeros((n_sel_events,), dtype=np.double)

        (src_idxs, evt_idxs) = tdm.src_evt_idxs
        for k in range(n_sources):
            src_mask = src_idxs == k
            R_i[evt_idxs[src_mask]] += R_ik[src_mask] * a_k[k]
        R_i /= A

        self._cache_R_ik = R_ik
        self._cache_R_i = R_i

        return R_i

    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None):
        """Retrieves the PDF ratio gradient for the parameter ``fitparam_id``
        for each trial data event, given the given set of parameters
        ``src_params_recarray`` for each source.

        Note:

            This method requires that the get_ratio method has been called prior
            to calling this method.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio gradient values should get calculated.
        src_params_recarray : instance of numpy record ndarray
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        fitparam_id : int
            The ID of the global fit parameter for which the gradient should
            get calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        gradient : instance of ndarray | 0
            The (N_selected_events,)-shaped 1d numpy ndarray of float holding
            the PDF ratio gradient value for each trial data event. If the PDF
            ratio does not depend on the given global fit parameter, 0 will
            be returned.
        """
        (a_jk, a_jk_grads) = self._src_detsigyield_weights_service.get_weights()

        a_k = a_jk[self._dataset_idx]
        A = np.sum(a_k)

        n_sources = a_jk.shape[1]
        n_sel_events = tdm.n_selected_events

        if fitparam_id not in a_jk_grads:
            a_k_grad = 0
            dAdp = 0
        else:
            a_k_grad = a_jk_grads[fitparam_id][self._dataset_idx]
            dAdp = np.sum(a_k_grad)

        R_ik_grad = self._pdfratio.get_gradient(
            tdm=tdm,
            src_params_recarray=src_params_recarray,
            fitparam_id=fitparam_id)
        # R_ik_grad is a (N_values,)-shaped ndarray or 0.

        if (type(a_k_grad) == int) and (a_k_grad == 0) and\
           (type(R_ik_grad) == int) and (R_ik_grad == 0):
            return 0

        R_i_grad = -self._cache_R_i * dAdp

        src_sum_i = np.zeros((n_sel_events,), dtype=np.double)

        (src_idxs, evt_idxs) = tdm.src_evt_idxs
        for k in range(n_sources):
            src_mask = src_idxs == k
            src_evt_idxs = evt_idxs[src_mask]
            if isinstance(a_k_grad, np.ndarray):
                src_sum_i[src_evt_idxs] +=\
                    a_k_grad[k] * self._cache_R_ik[src_mask]
            if isinstance(R_ik_grad, np.ndarray):
                src_sum_i[src_evt_idxs] +=\
                    a_k[k] * R_ik_grad[src_mask]

        R_i_grad += src_sum_i
        R_i_grad /= A

        return R_i_grad


class SigOverBkgPDFRatio(
        PDFRatio):
    """This class implements a generic signal-over-background PDF ratio for a
    signal and a background PDF instance.
    It takes a signal PDF of type *pdf_type* and a background PDF of type
    *pdf_type* and calculates the PDF ratio.
    """
    def __init__(
            self,
            sig_pdf,
            bkg_pdf,
            same_axes=True,
            zero_bkg_ratio_value=1.,
            **kwargs):
        """Creates a new signal-over-background PDF ratio instance.

        Parameters
        ----------
        sig_pdf : class instance derived from `pdf_type`, IsSignalPDF
            The instance of the signal PDF.
        bkg_pdf : class instance derived from `pdf_type`, IsBackgroundPDF
            The instance of the background PDF.
        same_axes : bool
            Flag if the signal and background PDFs are supposed to have the
            same axes. Default is True.
        zero_bkg_ratio_value : float
            The value of the PDF ratio to take when the background PDF value
            is zero. This is to avoid division by zero. Default is 1.
        """
        super().__init__(
            sig_param_names=sig_pdf.param_set.params_name_list,
            bkg_param_names=bkg_pdf.param_set.params_name_list,
            **kwargs)

        self.sig_pdf = sig_pdf
        self.bkg_pdf = bkg_pdf

        # Check that the PDF axes ranges are the same for the signal and
        # background PDFs.
        if same_axes and not sig_pdf.axes.is_same_as(bkg_pdf.axes):
            raise ValueError(
                'The signal and background PDFs do not have the same axes!')

        self.zero_bkg_ratio_value = zero_bkg_ratio_value

        # Define cache member variables to calculate gradients efficiently.
        self._cache_sig_pd = None
        self._cache_bkg_pd = None
        self._cache_sig_grads = None
        self._cache_bkg_grads = None

    @property
    def sig_pdf(self):
        """The signal PDF object used to create the PDF ratio.
        """
        return self._sig_pdf

    @sig_pdf.setter
    def sig_pdf(self, pdf):
        if not isinstance(pdf, IsSignalPDF):
            raise TypeError(
                'The sig_pdf property must be an instance of IsSignalPDF! '
                f'Its type is "{classname(pdf)}".')
        self._sig_pdf = pdf

    @property
    def bkg_pdf(self):
        """The background PDF object used to create the PDF ratio.
        """
        return self._bkg_pdf

    @bkg_pdf.setter
    def bkg_pdf(self, pdf):
        if not isinstance(pdf, IsBackgroundPDF):
            raise TypeError(
                'The bkg_pdf property must be an instance of IsBackgroundPDF! '
                f'Its type is "{classname(pdf)}".')
        self._bkg_pdf = pdf

    @property
    def zero_bkg_ratio_value(self):
        """The value of the PDF ratio to take when the background PDF value
        is zero. This is to avoid division by zero.
        """
        return self._zero_bkg_ratio_value

    @zero_bkg_ratio_value.setter
    def zero_bkg_ratio_value(self, v):
        v = float_cast(
            v,
            'The zero_bkg_ratio_value must be castable to type float!')
        self._zero_bkg_ratio_value = v

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Initializes the PDFRatio instance for a new trial. It calls the
        :meth:`~skyllh.core.pdf.PDF.assert_is_valid_for_trial_data` of the
        signal and background :class:`~skyllh.core.pdf.PDF` instances.
        """
        self._sig_pdf.initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)
        self._sig_pdf.assert_is_valid_for_trial_data(
            tdm=tdm,
            tl=tl,
            **kwargs)

        self._bkg_pdf.initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)
        self._bkg_pdf.assert_is_valid_for_trial_data(
            tdm=tdm,
            tl=tl,
            **kwargs)

    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None):
        """Calculates the PDF ratio for the given trial events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should be calculated.
        src_params_recarray : instance of numpy record ndarray
            The (N_sources,)-shaped numpy record ndarray holding the local
            parameter names and values of the sources.
            See the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            ratio for each event and source.
        """
        with TaskTimer(tl, 'Get sig probability densities and grads.'):
            (self._cache_sig_pd, self._cache_sig_grads) = self._sig_pdf.get_pd(
                tdm=tdm,
                params_recarray=src_params_recarray,
                tl=tl)
        with TaskTimer(tl, 'Get bkg probability densities and grads.'):
            (self._cache_bkg_pd, self._cache_bkg_grads) = self._bkg_pdf.get_pd(
                tdm=tdm,
                params_recarray=None,
                tl=tl)

        with TaskTimer(tl, 'Calculate PDF ratios.'):
            # Select only the events, where the background pdf is greater than
            # zero.
            ratios = np.full_like(self._cache_sig_pd, self._zero_bkg_ratio_value)
            m = (self._cache_bkg_pd > 0)
            (m, bkg_pd) = tdm.broadcast_selected_events_arrays_to_values_arrays(
                (m, self._cache_bkg_pd))
            np.divide(
                self._cache_sig_pd,
                bkg_pd,
                where=m,
                out=ratios)

        return ratios

    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None):
        """Retrieves the gradient of the PDF ratio w.r.t. the given parameter.

        Note:

            This method uses cached values from the
            :meth:`~skyllh.core.pdfratio.SigOverBkgPDFRatio.get_ratio` method.
            Hence, that method needs to be called prior to this method.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data.
        src_params_recarray : instance of numpy record ndarray | None
            The (N_models,)-shaped numpy record ndarray holding the parameter
            names and values of the models.
            See :meth:`skyllh.core.pdf.PDF.get_pd` for more information.
            This can be ``None``, if the signal and background PDFs do not
            depend on any parameters.
        fitparam_id : int
            The ID of the global fit parameter for which the gradient should
            get calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        grad : instance of ndarray
            The (N_values,)-shaped 1d numpy ndarray of float holding the PDF
            ratio gradient value for each source and trial event.
        """
        # Create the 1D return array for the gradient.
        grad = np.zeros_like(self._cache_sig_pd, dtype=np.float64)

        # Calculate the gradient for the given parameter.
        # There are four cases:
        #   1) Neither the signal nor the background PDF depend on the
        #      parameter.
        #   2) Only the signal PDF depends on the parameter.
        #   3) Only the background PDF depends on the parameter.
        #   4) Both, the signal and the background PDF depend on the
        #      parameter.
        sig_dep = fitparam_id in self._cache_sig_grads
        bkg_dep = fitparam_id in self._cache_bkg_grads

        if (not sig_dep) and (not bkg_dep):
            # Case 1. Return zeros.
            return grad

        m = self._cache_bkg_pd > 0
        b = self._cache_bkg_pd

        (m, b) = tdm.broadcast_selected_events_arrays_to_values_arrays(
            (m, b))

        if sig_dep and not bkg_dep:
            # Case 2, which should be the most common case.
            grad[m] = self._cache_sig_grads[fitparam_id][m] / b[m]
            return grad

        bgrad = self._cache_bkg_grads[fitparam_id]
        (bgrad,) = tdm.broadcast_selected_events_arrays_to_values_arrays(
            (bgrad,))

        if sig_dep and bkg_dep:
            # Case 4.
            s = self._cache_sig_pd
            sgrad = self._cache_sig_grads[fitparam_id]

            # Make use of quotient rule of differentiation.
            grad[m] = (sgrad[m] * b[m] - bgrad[m] * s[m]) / b[m]**2
            return grad

        # Case 3.
        grad[m] = -self._cache_sig_pd[m] / b[m]**2 * bgrad[m]

        return grad


class SigSetOverBkgPDFRatio(
        PDFRatio):
    """Class for a PDF ratio class that takes a PDFSet as signal PDF and a PDF
    as background PDF.
    The signal PDF depends on signal parameters and an interpolation method
    defines how the PDF ratio gets interpolated between the parameter grid
    values.
    """
    def __init__(
            self,
            sig_pdf_set,
            bkg_pdf,
            interpolmethod_cls=None,
            **kwargs):
        """Constructor called by creating an instance of a class which is
        derived from this PDFRatio class.

        Parameters
        ----------
        sig_pdf_set : instance of PDFSet and instance of IsSignalPDF
            The PDF set, which provides signal PDFs for a set of
            discrete signal parameter values.
        bkg_pdf : instance of PDF and instance of IsBackgroundPDF
            The background PDF instance.
        interpolmethod_cls : class of GridManifoldInterpolationMethod | None
            The class implementing the parameter interpolation method for
            the PDF ratio manifold grid. If set to ``None`` (default), the
            :class:`skyllh.core.interpolate.Parabola1DGridManifoldInterpolationMethod`
            will be used for 1-dimensional parameter manifolds.
        """
        super().__init__(
            sig_param_names=sig_pdf_set.param_grid_set.params_name_list,
            bkg_param_names=bkg_pdf.param_set.params_name_list,
            **kwargs)

        self.sig_pdf_set = sig_pdf_set
        self.bkg_pdf = bkg_pdf

        # Define the default parameter interpolation method. The default
        # depends on the dimensionality of the parameter manifold.
        if interpolmethod_cls is None:
            ndim = self._sig_pdf_set.param_grid_set.ndim
            if ndim == 1:
                interpolmethod_cls = Parabola1DGridManifoldInterpolationMethod
            else:
                raise ValueError(
                    'There is no default parameter manifold grid '
                    f'interpolation method class available for {ndim} '
                    'dimensions!')
        self.interpolmethod_cls = interpolmethod_cls

    @property
    def bkg_pdf(self):
        """The background PDF instance, derived from IsBackgroundPDF.
        """
        return self._bkg_pdf

    @bkg_pdf.setter
    def bkg_pdf(self, pdf):
        if not isinstance(pdf, IsBackgroundPDF):
            raise TypeError(
                'The bkg_pdf property must be an instance derived from '
                'IsBackgroundPDF! '
                f'Its current type is {classname(pdf)}.')
        self._bkg_pdf = pdf

    @property
    def sig_pdf_set(self):
        """The signal PDFSet instance, derived from IsSignalPDF.
        """
        return self._sig_pdf_set

    @sig_pdf_set.setter
    def sig_pdf_set(self, pdfset):
        if not (isinstance(pdfset, PDFSet) and
                isinstance(pdfset, IsSignalPDF)):
            raise TypeError(
                'The sig_pdf_set property must be a class instance which is '
                'derived from PDFSet and IsSignalPDF! '
                f'Its current type is {classname(pdfset)}.')
        self._sig_pdf_set = pdfset

    @property
    def interpolmethod_cls(self):
        """The class derived from GridManifoldInterpolationMethod
        implementing the interpolation of the parameter manifold.
        """
        return self._interpolmethod_cls

    @interpolmethod_cls.setter
    def interpolmethod_cls(self, cls):
        if not issubclass(cls, GridManifoldInterpolationMethod):
            raise TypeError(
                'The interpolmethod_cls property must be a sub-class '
                'of GridManifoldInterpolationMethod! '
                f'Its current type is {classname(cls)}.')
        self._interpolmethod_cls = cls

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Initializes the PDFRatio instance for a new trial. It calls the
        :meth:`~skyllh.core.pdf.PDF.assert_is_valid_for_trial_data` of the
        signal :class:`~skyllh.core.pdf.PDFSet` instance and the background
        :class:`~skyllh.core.pdf.PDF` instance.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data.
        tl : instance of TimeLord | None
            The optional instance of TimeLord for measuring timing information.
        """
        self._sig_pdf_set.initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)
        self._sig_pdf_set.assert_is_valid_for_trial_data(
            tdm=tdm,
            tl=tl,
            **kwargs)

        self._bkg_pdf.initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)
        self._bkg_pdf.assert_is_valid_for_trial_data(
            tdm=tdm,
            tl=tl,
            **kwargs)



# -*- coding: utf-8 -*-

import os.path

import numpy as np

from skyllh.core import (
    display,
)
from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.py import (
    issequenceof,
    module_classname,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
    create_FileLoader,
)
from skyllh.core.timing import (
    TaskTimer,
)


class I3Dataset(
        Dataset,
):
    """The I3Dataset class is an IceCube specific Dataset class that adds
    IceCube specific properties to the Dataset class. These additional
    properties are:

        * good-run-list (GRL)

    """
    @staticmethod
    def get_combined_grl_pathfilenames(datasets):
        """Creates the combined list of grl pathfilenames of all the given
        datasets.

        Parameters
        ----------
        datasets : sequence of I3Dataset
            The sequence of I3Dataset instances.

        Returns
        -------
        grl_pathfilenames : list
            The combined list of grl pathfilenames.
        """
        if not issequenceof(datasets, I3Dataset):
            raise TypeError(
                'The datasets argument must be a sequence of I3Dataset '
                'instances!')

        grl_pathfilenames = []
        for ds in datasets:
            grl_pathfilenames += ds.grl_pathfilename_list

        return grl_pathfilenames

    def __init__(
            self,
            livetime=None,
            grl_pathfilenames=None,
            **kwargs,
    ):
        """Creates a new IceCube specific dataset, that also can hold a list
        of GRL data files.

        Parameters
        ----------
        livetime : float | None
            The live-time of the dataset in days. It can be ``None``, if
            good-run-list data files are provided.
        grl_pathfilenames : str | sequence of str
            The sequence of pathfilenames pointing to the good-run-list (GRL)
            data files.
        """
        super().__init__(
            livetime=livetime,
            **kwargs)

        self._logger = get_logger(module_classname(self))

        self.grl_pathfilename_list = grl_pathfilenames

        self.grl_field_name_renaming_dict = dict()

    @property
    def grl_pathfilename_list(self):
        """The list of file names of the good-run-list (GRL) data files for this
        dataset. If a file name is given with a relative path, it will be
        relative to the ``root_dir`` property of this Dataset instance.
        """
        return self._grl_pathfilename_list

    @grl_pathfilename_list.setter
    def grl_pathfilename_list(self, pathfilenames):
        if pathfilenames is None:
            pathfilenames = []
        if isinstance(pathfilenames, str):
            pathfilenames = [pathfilenames]
        if not issequenceof(pathfilenames, str):
            raise TypeError(
                'The grl_pathfilename_list property must be a sequence of str!')
        self._grl_pathfilename_list = list(pathfilenames)

    @property
    def grl_abs_pathfilename_list(self):
        """(read-only) The list of absolute path file names of the good-run-list
        data files.
        """
        return self.get_abs_pathfilename_list(self._grl_pathfilename_list)

    @property
    def grl_field_name_renaming_dict(self):
        """The dictionary specifying the field names of the good-run-list data
        which need to get renamed just after loading the data. The dictionary
        keys are the old names and their values are the new names.
        """
        return self._grl_field_name_renaming_dict

    @grl_field_name_renaming_dict.setter
    def grl_field_name_renaming_dict(self, d):
        if not isinstance(d, dict):
            raise TypeError(
                'The grl_field_name_renaming_dict property must be an '
                'instance of dict!')
        self._grl_field_name_renaming_dict = d

    @property
    def exists(self):
        """(read-only) Flag if all the data files of this data set exists. It is
        ``True`` if all data files exist and ``False`` otherwise.
        """
        if not super().exists:
            return False

        for pathfilename in self.grl_abs_pathfilename_list:
            if not os.path.exists(pathfilename):
                return False

        return True

    def __str__(self):
        """Implementation of the pretty string representation of the I3Dataset
        object.
        """
        s = super().__str__()
        s += '\n'

        s1 = ''
        s1 += 'GRL data:\n'
        s2 = ''
        if len(self._grl_pathfilename_list) > 0:
            for (idx, pathfilename) in enumerate(self.grl_abs_pathfilename_list):
                if idx > 0:
                    s2 += '\n'
                s2 += self._gen_datafile_pathfilename_entry(pathfilename)
        else:
            s2 += 'None'
        s1 += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s2)

        s += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s1)

        return s

    def create_file_list(
            self,
    ):
        """Creates the list of files of this dataset.
        The file paths are relative to the dataset's root directory.

        Returns
        -------
        file_list : list of str
            The list of files of this dataset.
        """
        file_list = (
            super().create_file_list() +
            self._grl_pathfilename_list
        )

        return file_list

    def load_grl(
            self,
            efficiency_mode=None,
            tl=None,
    ):
        """Loads the good-run-list and returns a DataFieldRecordArray instance
        which should contain the following data fields:

            run : int
                The run number.
            start : float
                The MJD start time of the run.
            stop : float
                The MJD stop time of the run.
            livetime : float
                The livetime in days of the run.
            events : int
                The number of experimental events in the run.

        Parameters
        ----------
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                - 'memory':
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                - 'time'
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.
        tl : TimeLord instance | None
            The TimeLord instance to use to time the data loading procedure.

        Returns
        -------
        grl_data : instance of DataFieldRecordArray
            The DataFieldRecordArray instance holding the good-run-list
            information of the dataset.
        """
        with TaskTimer(tl, 'Loading grl data from disk.'):
            fileloader_grl = create_FileLoader(
                self.grl_abs_pathfilename_list)
            grl_data = fileloader_grl.load_data(
                efficiency_mode=efficiency_mode)
            grl_data.rename_fields(self._grl_field_name_renaming_dict)

        with TaskTimer(tl, 'Sort grl data according to start time'):
            grl_data.sort_by_field(name='start')

        return grl_data

    def load_data(
            self,
            livetime=None,
            keep_fields=None,
            dtc_dict=None,
            dtc_except_fields=None,
            efficiency_mode=None,
            tl=None,
    ):
        """Loads the data, which is described by the dataset. If a good-run-list
        (GRL) is provided for this dataset, only experimental data will be
        selected which matches the GRL.

        Parameters
        ----------
        livetime : instance of Livetime | float | None
            If not None, uses this livetime (if float livetime in days) as
            livetime for the DatasetData instance, otherwise uses the live time
            from the Dataset instance or, if available, the livetime from the
            good-run-list (GRL).
        keep_fields : list of str | None
            The list of user-defined data fields that should get loaded and kept
            in addition to the analysis required data fields.
        dtc_dict : dict | None
            This dictionary defines how data fields of specific data types (key)
            should get converted into other data types (value).
            This can be used to use less memory. If set to None, no data
            convertion is performed.
        dtc_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                - 'memory':
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                - 'time'
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.
        tl : TimeLord instance | None
            The TimeLord instance that should be used to time the data load
            operation.

        Returns
        -------
        data : instance of DatasetData
            A DatasetData instance holding the experimental and monte-carlo
            data of this data set.
        """
        # Load the dataset files first. This will ensure the dataset is
        # downloaded if necessary.
        data_ = super().load_data(
            livetime=livetime,
            keep_fields=keep_fields,
            dtc_dict=dtc_dict,
            dtc_except_fields=dtc_except_fields,
            efficiency_mode=efficiency_mode,
            tl=tl)

        # Load the good-run-list (GRL) data if it is provided for this dataset,
        # and calculate the livetime based on the GRL.
        data_grl = None
        if len(self._grl_pathfilename_list) > 0:
            data_grl = self.load_grl(
                efficiency_mode=efficiency_mode,
                tl=tl)

        # Load all the defined data.
        data = I3DatasetData(
            data=data_,
            data_grl=data_grl)

        return data

    def prepare_data(  # noqa: C901
            self,
            data,
            tl=None,
    ):
        """Prepares the data for IceCube by pre-calculating the following
        experimental data fields:

        - sin_dec: float
            The sin value of the declination coordinate.

        and monte-carlo data fields:

        - sin_true_dec: float
            The sin value of the true declination coordinate.

        Parameters
        ----------
        data : DatasetData instance
            The DatasetData instance holding the data as numpy record ndarray.
        tl : TimeLord instance | None
            The TimeLord instance that should be used to time the data
            preparation.
        """
        # Set the livetime of the dataset from the GRL data when no livetime
        # was specified previously.
        if data.livetime is None and data.grl is not None:
            if 'livetime' in data.grl:
                # The `livetime` column accounts for livetime gaps within run,
                # therefore it is more precise than `stop` - `start`.
                data.livetime = np.sum(data.grl['livetime'])
            else:
                # If `livetime` field does not exist fall back to
                # `stop` - `start` implementation.
                if 'start' not in data.grl:
                    raise KeyError(
                        f'The GRL data for dataset "{self.name}" has no data '
                        'field named "start"!')
                if 'stop' not in data.grl:
                    raise KeyError(
                        f'The GRL data for dataset "{self.name}" has no data '
                        'field named "stop"!')
                data.livetime = np.sum(data.grl['stop'] - data.grl['start'])

        # Execute all the data preparation functions for this dataset.
        super().prepare_data(
            data=data,
            tl=tl)

        if data.exp is not None:
            # Append sin(dec) data field to the experimental data.
            task = 'Appending IceCube-specific data fields to exp data.'
            with TaskTimer(tl, task):
                if 'sin_dec' not in data.exp.field_name_list:
                    data.exp.append_field(
                        'sin_dec', np.sin(data.exp['dec']))

        if data.mc is not None:
            # Append sin(dec) and sin(true_dec) to the MC data.
            task = 'Appending IceCube-specific data fields to MC data.'
            with TaskTimer(tl, task):
                if 'sin_dec' not in data.mc.field_name_list:
                    data.mc.append_field(
                        'sin_dec', np.sin(data.mc['dec']))
                if 'sin_true_dec' not in data.mc.field_name_list:
                    data.mc.append_field(
                        'sin_true_dec', np.sin(data.mc['true_dec']))

        # Select only the experimental data which fits the good-run-list for
        # this dataset.
        if (data.grl is not None) and (data.exp is not None):
            # Select based on run information.
            if ('run' in data.grl) and ('run' in data.exp):
                task = (
                    'Select only the experimental data that matches the run '
                    f'information in the GRL for dataset "{self.name}".')
                with TaskTimer(tl, task):
                    runs = np.unique(data.grl['run'])
                    mask = np.isin(data.exp['run'], runs)

                    if np.any(~mask):
                        n_cut_runs = np.count_nonzero(~mask)
                        self._logger.info(
                            f'Cutting {n_cut_runs} runs from dataset '
                            f'{self.name} due to GRL run information.')
                        data.exp = data.exp[mask]

            # Select based on detector on-time information.
            if ('start' in data.grl) and\
               ('stop' in data.grl) and\
               ('time' in data.exp):
                task = (
                    'Select only the experimental data that matches the '
                    'detector\'s on-time information in the GRL for dataset '
                    f'"{self.name}".')
                with TaskTimer(tl, task):
                    mask = np.zeros((len(data.exp),), dtype=np.bool_)
                    for (start, stop) in zip(data.grl['start'],
                                             data.grl['stop']):
                        mask |= (
                            (data.exp['time'] >= start) &
                            (data.exp['time'] <= stop)
                        )

                    if np.any(~mask):
                        n_cut_evts = np.count_nonzero(~mask)
                        self._logger.info(
                            f'Cutting {n_cut_evts} events from dataset '
                            f'{self.name} due to GRL on-time window '
                            'information.')
                        data.exp = data.exp[mask]


class I3DatasetData(
        DatasetData,
):
    """The class provides the container for the loaded experimental and
    monto-carlo data of a data set. It's the IceCube specific class that also
    holds the good-run-list (GRL) data.
    """
    def __init__(
            self,
            data,
            data_grl,
    ):
        """Constructs a new I3DatasetData instance.

        Parameters
        ----------
        data : instance of DatasetData
            The instance of DatasetData holding the experimental and monte-carlo
            data.
        data_grl : instance of DataFieldRecordArray | None
            The instance of DataFieldRecordArray holding the good-run-list data
            of the dataset. This can be None, if no GRL data is available.
        """
        super().__init__(
            data_exp=data._exp,
            data_mc=data._mc,
            livetime=data._livetime)

        self.grl = data_grl

    @property
    def grl(self):
        """The DataFieldRecordArray instance holding the good-run-list (GRL)
        data of the IceCube data set. It is None, if there is no GRL data
        available for this IceCube data set.
        """
        return self._grl

    @grl.setter
    def grl(self, data):
        if data is not None:
            if not isinstance(data, DataFieldRecordArray):
                raise TypeError(
                    'The grl property must be an instance of '
                    'DataFieldRecordArray!')
        self._grl = data



# -*- coding: utf-8 -*-

"""This modules defines base types for some of the SkyLLH classes to avoid
circular imports when actively checking for types.
"""


class SourceHypoGroup_t(
    object,
):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

"""The mcbkg_ps analysis is a multi-dataset time-integrated single source
analysis with a two-component likelihood function using a spacial and an energy
event PDF. It initializes the background energy pdf using auxiliary fluxes and
pdfs, which are generated by running `scripts/mceq_atm_bkg.py` script.
"""

import argparse
import logging
import numpy as np
import pickle

from skyllh.analyses.i3.publicdata_ps.backgroundpdf import (
    PDMCBackgroundI3EnergyPDF,
)
from skyllh.analyses.i3.publicdata_ps.detsigyield import (
    PDSingleParamFluxPointLikeSourceI3DetSigYieldBuilder,
)
from skyllh.analyses.i3.publicdata_ps.pdfratio import (
    PDSigSetOverBkgPDFRatio,
)
from skyllh.analyses.i3.publicdata_ps.signal_generator import (
    PDDatasetSignalGenerator,
)
from skyllh.analyses.i3.publicdata_ps.signalpdf import (
    PDSignalEnergyPDFSet,
)
from skyllh.analyses.i3.publicdata_ps.utils import (
    create_energy_cut_spline,
)
from skyllh.core.analysis import (
    SingleSourceMultiDatasetLLHRatioAnalysis as Analysis,
)
from skyllh.core.config import (
    Config,
)
from skyllh.core.debugging import (
    get_logger,
    setup_logger,
    setup_console_handler,
    setup_file_handler,
)
from skyllh.core.event_selection import (
    SpatialBoxEventSelectionMethod,
)
from skyllh.core.flux_model import (
    PowerLawEnergyFluxProfile,
    SteadyPointlikeFFM,
)
from skyllh.core.minimizer import (
    LBFGSMinimizerImpl,
    Minimizer,
)
from skyllh.core.minimizers.iminuit import (
    IMinuitMinimizerImpl,
)
from skyllh.core.model import (
    DetectorModel,
)
from skyllh.core.parameters import (
    Parameter,
    ParameterModelMapper,
)
from skyllh.core.pdfratio import (
    SigOverBkgPDFRatio,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.random import (
    RandomStateService,
)
from skyllh.core.scrambling import (
    DataScrambler,
    UniformRAScramblingMethod,
)
from skyllh.core.signal_generator import (
    MultiDatasetSignalGenerator,
)
from skyllh.core.signalpdf import (
    RayleighPSFPointSourceSignalSpatialPDF,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroup,
    SourceHypoGroupManager,
)
from skyllh.core.source_model import (
    PointLikeSource,
)
from skyllh.core.test_statistic import (
    WilksTestStatistic,
)
from skyllh.core.timing import (
    TimeLord,
)
from skyllh.core.trialdata import (
    TrialDataManager,
)
from skyllh.core.utils.analysis import (
    pointlikesource_to_data_field_array,
)
from skyllh.core.utils.tdm import (
    get_tdm_field_func_psi,
)
from skyllh.datasets.i3 import (
    data_samples,
)
from skyllh.i3.background_generation import (
    FixedScrambledExpDataI3BkgGenMethod,
)
from skyllh.i3.backgroundpdf import (
    DataBackgroundI3SpatialPDF,
)


def TXS_location():
    src_ra = np.radians(77.358)
    src_dec = np.radians(5.693)
    return (src_ra, src_dec)


def create_analysis(
    datasets,
    source,
    refplflux_Phi0=1,
    refplflux_E0=1e3,
    refplflux_gamma=2,
    ns_seed=100,
    ns_min=0,
    ns_max=1e3,
    gamma_seed=3,
    gamma_min=1,
    gamma_max=5,
    minimizer_impl='LBFGS',
    cut_sindec=None,
    spl_smooth=None,
    cap_ratio=False,
    compress_data=False,
    keep_data_fields=None,
    evt_sel_delta_angle_deg=10,
    efficiency_mode=None,
    tl=None,
    ppbar=None,
    logger_name=None,
):
    """Creates the Analysis instance for this particular analysis.

    Parameters
    ----------
    datasets : list of Dataset instances
        The list of Dataset instances, which should be used in the
        analysis.
    source : PointLikeSource instance
        The PointLikeSource instance defining the point source position.
    refplflux_Phi0 : float
        The flux normalization to use for the reference power law flux model.
    refplflux_E0 : float
        The reference energy to use for the reference power law flux model.
    refplflux_gamma : float
        The spectral index to use for the reference power law flux model.
    ns_seed : float
        Value to seed the minimizer with for the ns fit.
    ns_min : float
        Lower bound for ns fit.
    ns_max : float
        Upper bound for ns fit.
    gamma_seed : float | None
        Value to seed the minimizer with for the gamma fit. If set to None,
        the refplflux_gamma value will be set as gamma_seed.
    gamma_min : float
        Lower bound for gamma fit.
    gamma_max : float
        Upper bound for gamma fit.
    minimizer_impl : str
        Minimizer implementation to be used. Supported options are ``"LBFGS"``
        (L-BFG-S minimizer used from the :mod:`scipy.optimize` module), or
        ``"minuit"`` (Minuit minimizer used by the :mod:`iminuit` module).
        Default: "LBFGS".
    cut_sindec : list of float | None
        sin(dec) values at which the energy cut in the southern sky should
        start. If None, np.sin(np.radians([-2, 0, -3, 0, 0])) is used.
    spl_smooth : list of float
        Smoothing parameters for the 1D spline for the energy cut. If None,
        [0., 0.005, 0.05, 0.2, 0.3] is used.
    compress_data : bool
        Flag if the data should get converted from float64 into float32.
    keep_data_fields : list of str | None
        List of additional data field names that should get kept when loading
        the data.
    evt_sel_delta_angle_deg : float
        The delta angle in degrees for the event selection optimization methods.
    efficiency_mode : str | None
        The efficiency mode the data should get loaded with. Possible values
        are:

        ``'memory'``:
            The data will be load in a memory efficient way. This will
            require more time, because all data records of a file will
            be loaded sequentially.
        ``'time'``:
            The data will be loaded in a time efficient way. This will
            require more memory, because each data file gets loaded in
            memory at once.

        The default value is ``'time'``. If set to ``None``, the default
        value will be used.
    tl : TimeLord instance | None
        The TimeLord instance to use to time the creation of the analysis.
    ppbar : ProgressBar instance | None
        The instance of ProgressBar for the optional parent progress bar.
    logger_name : str | None
        The name of the logger to be used. If set to ``None``, ``__name__`` will
        be used.

    Returns
    -------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The Analysis instance for this analysis.
    """
    if logger_name is None:
        logger_name = __name__
    logger = get_logger(logger_name)

    # Create the minimizer instance.
    if minimizer_impl == 'LBFGS':
        minimizer = Minimizer(LBFGSMinimizerImpl())
    elif minimizer_impl == 'minuit':
        minimizer = Minimizer(IMinuitMinimizerImpl(ftol=1e-8))
    else:
        raise NameError(
            f"Minimizer implementation `{minimizer_impl}` is not supported "
            "Please use `LBFGS` or `minuit`.")

    # Define the flux model.
    fluxmodel = SteadyPointlikeFFM(
        Phi0=refplflux_Phi0,
        energy_profile=PowerLawEnergyFluxProfile(
            E0=refplflux_E0,
            gamma=refplflux_gamma))

    # Define the fit parameter ns.
    param_ns = Parameter(
        name='ns',
        initial=ns_seed,
        valmin=ns_min,
        valmax=ns_max)

    # Define the fit parameter gamma.
    param_gamma = Parameter(
        name='gamma',
        initial=gamma_seed,
        valmin=gamma_min,
        valmax=gamma_max)

    # Define the detector signal efficiency implementation method for the
    # IceCube detector and this source and flux_model.
    # The sin(dec) binning will be taken by the implementation method
    # automatically from the Dataset instance.
    gamma_grid = param_gamma.as_linear_grid(delta=0.1)
    detsigyield_builder =\
        PDSingleParamFluxPointLikeSourceI3DetSigYieldBuilder(
            param_grid=gamma_grid)

    # Create a source hypothesis group manager.
    shg_mgr = SourceHypoGroupManager(
        SourceHypoGroup(
            sources=source,
            fluxmodel=fluxmodel,
            detsigyield_builders=detsigyield_builder,
            sig_gen_method=None))

    # Define a detector model for the ns fit parameter.
    detector_model = DetectorModel('IceCube')

    # Define the parameter model mapper for the analysis, which will map global
    # parameters to local source parameters.
    pmm = ParameterModelMapper(
        models=[detector_model, source])
    pmm.map_param(param_ns, models=detector_model)
    pmm.map_param(param_gamma, models=source)

    logger.info(str(pmm))

    # Define the test statistic.
    test_statistic = WilksTestStatistic()

    # Define the data scrambler with its data scrambling method, which is used
    # for background generation.
    data_scrambler = DataScrambler(UniformRAScramblingMethod())

    # Create background generation method.
    bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(data_scrambler)

    # Create the minimizer instance.
    minimizer = Minimizer(LBFGSMinimizerImpl())

    # Create the Analysis instance.
    ana = Analysis(
        shg_mgr=shg_mgr,
        pmm=pmm,
        test_statistic=test_statistic,
        bkg_gen_method=bkg_gen_method,
        sig_generator_cls=MultiDatasetSignalGenerator,
    )

    # Define the event selection method for pure optimization purposes.
    # We will use the same method for all datasets.
    event_selection_method = SpatialBoxEventSelectionMethod(
        shg_mgr=shg_mgr,
        delta_angle=np.deg2rad(evt_sel_delta_angle_deg))

    # Prepare the spline parameters for the signal generator.
    if cut_sindec is None:
        cut_sindec = np.sin(np.radians([-2, 0, -3, 0, 0]))
    if spl_smooth is None:
        spl_smooth = [0., 0.005, 0.05, 0.2, 0.3]
    if len(spl_smooth) < len(datasets) or len(cut_sindec) < len(datasets):
        raise AssertionError(
            'The length of the spl_smooth and of the cut_sindec must be equal '
            f'to the length of datasets: {len(datasets)}.')

    # Add the data sets to the analysis.
    pbar = ProgressBar(len(datasets), parent=ppbar).start()
    for (ds_idx, ds) in enumerate(datasets):
        # Load the data of the data set.
        data = ds.load_and_prepare_data(
            keep_fields=keep_data_fields,
            compress=compress_data,
            efficiency_mode=efficiency_mode,
            tl=tl)

        sin_dec_binning = ds.get_binning_definition('sin_dec')

        # Create the spatial PDF ratio instance for this dataset.
        spatial_sigpdf = RayleighPSFPointSourceSignalSpatialPDF(
            dec_range=np.arcsin(sin_dec_binning.range))
        spatial_bkgpdf = DataBackgroundI3SpatialPDF(
            data_exp=data.exp,
            sin_dec_binning=sin_dec_binning)
        spatial_pdfratio = SigOverBkgPDFRatio(
            sig_pdf=spatial_sigpdf,
            bkg_pdf=spatial_bkgpdf)

        # Create the energy PDF ratio instance for this dataset.
        energy_sigpdfset = PDSignalEnergyPDFSet(
            ds=ds,
            src_dec=source.dec,
            fluxmodel=fluxmodel,
            param_grid_set=gamma_grid,
            ppbar=ppbar
        )

        bkg_pdf_pathfilename = ds.get_abs_pathfilename_list(
            ds.get_aux_data_definition('pdf_bkg_datafile'))[0]
        with open(bkg_pdf_pathfilename, 'rb') as f:
            bkg_pdf_data = pickle.load(f)
        energy_bkgpdf = PDMCBackgroundI3EnergyPDF(
            pdf_log10emu_sindecmu=bkg_pdf_data['pdf'],
            log10emu_binning=bkg_pdf_data['log10emu_binning'],
            sindecmu_binning=bkg_pdf_data['sindecmu_binning'],
        )

        energy_pdfratio = PDSigSetOverBkgPDFRatio(
            sig_pdf_set=energy_sigpdfset,
            bkg_pdf=energy_bkgpdf,
            cap_ratio=cap_ratio)

        pdfratio = spatial_pdfratio * energy_pdfratio

        # Create a trial data manager and add the required data fields.
        tdm = TrialDataManager()
        tdm.add_source_data_field(
            name='src_array',
            func=pointlikesource_to_data_field_array)
        tdm.add_data_field(
            name='psi',
            func=get_tdm_field_func_psi(),
            dt='dec',
            is_srcevt_data=True)

        energy_cut_spline = create_energy_cut_spline(
            ds,
            data.exp,
            spl_smooth[ds_idx])

        sig_generator = PDDatasetSignalGenerator(
            shg_mgr=shg_mgr,
            ds=ds,
            ds_idx=ds_idx,
            energy_cut_spline=energy_cut_spline,
            cut_sindec=cut_sindec[ds_idx],
        )

        ana.add_dataset(
            dataset=ds,
            data=data,
            pdfratio=pdfratio,
            tdm=tdm,
            event_selection_method=event_selection_method,
            sig_generator=sig_generator)

        pbar.increment()
    pbar.finish()

    ana.construct_services(
        ppbar=ppbar)

    ana.llhratio = ana.construct_llhratio(
        minimizer=minimizer,
        ppbar=ppbar)

    ana.construct_signal_generator()

    return ana


if __name__ == '__main__':
    p = argparse.ArgumentParser(
        description='Calculates TS for a given source location using the '
        '10-year public point source sample.',
        formatter_class=argparse.RawTextHelpFormatter
    )
    p.add_argument(
        '--dec',
        default=23.8,
        type=float,
        help='The source declination in degrees.'
    )
    p.add_argument(
        '--ra',
        default=216.76,
        type=float,
        help='The source right-ascention in degrees.'
    )
    p.add_argument(
        '--gamma-seed',
        default=3,
        type=float,
        help='The seed value of the gamma fit parameter.'
    )
    p.add_argument(
        '--data_base_path',
        default=None,
        type=str,
        help='The base path to the data samples (default=None)'
    )
    p.add_argument(
        '--seed',
        default=1,
        type=int,
        help='The random number generator seed for the likelihood '
             'minimization.'
    )
    p.add_argument(
        '--ncpu',
        default=1,
        type=int,
        help='The number of CPUs to utilize where parallelization is possible.'
    )
    p.add_argument(
        '--cap-ratio',
        action='store_true',
        help='Switch to cap the energy PDF ratio.')
    p.set_defaults(cap_ratio=False)
    args = p.parse_args()

    # Setup `skyllh` package logging.
    # To optimize logging set the logging level to the lowest handling level.
    setup_logger('skyllh', logging.DEBUG)
    log_format = '%(asctime)s %(processName)s %(name)s %(levelname)s: '\
                 '%(message)s'
    setup_console_handler(
        'skyllh',
        logging.INFO, log_format)
    setup_file_handler(
        'skyllh',
        'debug.log',
        log_level=logging.DEBUG,
        log_format=log_format)

    cfg = Config()
    cfg.set_ncpu(args.ncpu)

    sample_seasons = [
        # ('PublicData_10y_ps', 'IC40'),
        # ('PublicData_10y_ps', 'IC59'),
        # ('PublicData_10y_ps', 'IC79'),
        # ('PublicData_10y_ps', 'IC86_I'),
        ('PublicData_10y_ps', 'IC86_II'),
        # ('PublicData_10y_ps', 'IC86_II-VII')
    ]

    datasets = []
    for (sample, season) in sample_seasons:
        # Get the dataset from the correct dataset collection.
        dsc = data_samples[sample].create_dataset_collection(
            args.data_base_path)
        datasets.append(dsc.get_dataset(season))

    # Define a random state service.
    rss = RandomStateService(args.seed)
    # Define the point source.
    source = PointLikeSource(np.deg2rad(args.ra), np.deg2rad(args.dec))
    print('source: ', str(source))

    tl = TimeLord()

    with tl.task_timer('Creating analysis.'):
        ana = create_analysis(
            datasets=datasets,
            source=source,
            cap_ratio=args.cap_ratio,
            gamma_seed=args.gamma_seed,
            tl=tl)

    with tl.task_timer('Unblinding data.'):
        (TS, fitparam_dict, status) = ana.unblind(rss)

    print('TS = %g' % (TS))
    print('ns_fit = %g' % (fitparam_dict['ns']))
    print('gamma_fit = %g' % (fitparam_dict['gamma']))

    print(tl)
